from django.core.exceptions import ValidationError
from django.db import models
from django.db.utils import IntegrityError


class AssignAttributes:
    def save(self, instance, bundle):
        items = [
            (key, getattr(self.model, key), value) for key, value in bundle.items()
        ]

        for key, attr, value in items:
            if isinstance(value, models.Model):
                setattr(instance, key, value)
                continue

            if isinstance(attr.field, models.ForeignKey):
                self.set_foreign_key(instance, key, value)
                continue

            if isinstance(attr.field, models.ManyToManyField):
                continue

            setattr(instance, key, value)

        instance.save()

        for key, attr, value in items:
            if isinstance(attr.field, models.ManyToManyField):
                has_through = not attr.through._meta.auto_created

                if has_through and any(isinstance(item, dict) for item in value):
                    self.set_many_to_many_with_through(instance, key, value)
                    continue

                self.set_many_to_many(instance, key, value)

    def set_foreign_key(self, instance, key, value):
        related_model = self.get_related_model(key)
        related_model_meta = getattr(related_model, "Api", None)
        lookup_field = getattr(related_model_meta, "lookup_field", "pk")
        if value is not None:
            try:
                value = related_model.objects.get(**{lookup_field: value})
            except related_model.DoesNotExist as e:
                raise ValidationError(f"Invalid {self.keymap[key]}") from e
        setattr(instance, key, value)

    def set_many_to_many(self, instance, key, value):
        related_manager = getattr(instance, key)
        related_model = related_manager.model
        related_model_meta = getattr(related_model, "Api", None)
        lookup_field = getattr(related_model_meta, "lookup_field", "pk")
        try:
            if lookup_field != "pk":
                results = related_model.objects.filter(**{f"{lookup_field}__in": value})
                assert len(results) == len(value)
                value = results
            related_manager.set(value)
        except (AssertionError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def set_many_to_many_with_through(self, instance, key, value):
        try:
            attr = getattr(self.model, key)

            through_model = attr.through
            model_name = self.model._meta.model_name
            target_field_name = attr.field.m2m_target_field_name()
            reverse_name = attr.field.m2m_reverse_name()

            getattr(instance, key).clear()

            through_model.objects.bulk_create(
                [
                    through_model(
                        **{
                            item_key: item_value
                            for item_key, item_value in item.items()
                            if item_key != target_field_name
                        },
                        **{
                            model_name: instance,
                            reverse_name: item[target_field_name],
                        },
                    )
                    for item in value
                ]
            )
        except (AttributeError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def validate(self):
        instance = self.get_instance()

        for key in self.bundle.keys():
            self.validate_bundle(key)

            field = self.model._meta.get_field(key)

            if self.bundle[key] is None and not field.null:
                raise ValidationError(f"Invalid {self.keymap[key]}")

            if field.unique:
                other_records = self.model.objects.exclude(pk=instance.pk)

                if other_records.filter(**{key: self.bundle[key]}).exists():
                    raise ValidationError(f"Field {self.keymap[key]} must be unique")
